// NpUdpSession.cpp: implementation of the NpUdpSession class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "NpUdpSession.h"

#include "NpPacketSender.h"
#include "NpSessionBase.h"

#define DGRAM_SEND_BUFFER_DEFAULT_LEN	8192

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CNpUdpSession::CNpUdpSession()
{	
	m_pIckarSocket = NULL;
	m_pEvents = NULL;
}

CNpUdpSession::~CNpUdpSession()
{	
	DLOG("CNpUdpSession::~CNpUdpSession() %p %p\n", this, m_pIckarSocket);

	if( m_pIckarSocket )
	{
		m_pIckarSocket->Delete();
	}
}

bool CNpUdpSession::InitIckarSocket()
{
	if( m_pIckarSocket )
	{
		m_pIckarSocket->Close();
	}
	else
	{
		m_pIckarSocket = _CreateIckarSocket(this);
	}
	
	return true;
}

void CNpUdpSession::SetEvents(_INpUdpSessionEvents * pEvents)
{
	m_pEvents = pEvents;
}

IIckarSocket * CNpUdpSession::GetIckarSocket()
{
	if( m_pIckarSocket == NULL )
		InitIckarSocket();
	
	return m_pIckarSocket;
}

bool CNpUdpSession::AttachIckarSocket(IIckarSocket * pIckarSocket)
{
	if( m_pIckarSocket )
	{
		m_pIckarSocket->Delete();
	}
	
	m_pIckarSocket = pIckarSocket;
	m_pIckarSocket->SetEvents( this );
	
	return true;
}

IIckarSocket * CNpUdpSession::DetachIckarSocket()
{
	if( m_pIckarSocket == NULL )
		return NULL;
	
	IIckarSocket * pIckarSocket;
	pIckarSocket = m_pIckarSocket;
	m_pIckarSocket = NULL;
	
	pIckarSocket->SetEvents( NULL );
	return pIckarSocket;
}

bool CNpUdpSession::ConnectSession(LPCSTR szHostAddr, int nHostPort, bool bUseProxySetting)
{
	if( m_pIckarSocket == NULL )
		InitIckarSocket();

	ASSERT( m_pIckarSocket->IsConnected() == false );

	return m_pIckarSocket->ConnectEx( szHostAddr, nHostPort, eNCS_PT_UDP );
}

bool CNpUdpSession::StartListen(int nBindPort, LPCSTR sBindAddr)
{
	if( m_pIckarSocket == NULL )
		InitIckarSocket();

	ASSERT( m_pIckarSocket->IsConnected() == false );

	return m_pIckarSocket->BindEx( nBindPort, sBindAddr, eNCS_PT_UDP );
}

void CNpUdpSession::CloseSession()
{
	if( m_pIckarSocket )
		m_pIckarSocket->Close();
}

bool CNpUdpSession::SendNpPacket(const INpPacket * pPacket, LPCSTR szTargetIP, int nTargetPort)
{
	DLOG("CNpUdpSession::SendNpPacket(pPacket=%p), packet header first byte=[%x]", (*pPacket->RawHeader()) & 0x000000FF );
	
	NPPACKET_DUMP( pPacket );
	
	if( IsConnectMode() )
	{
		ASSERT( (szTargetIP==NULL || strlen(szTargetIP) == 0 ) && nTargetPort == 0 );

		return CNpPacketSender::SendNpPacket( m_pIckarSocket, pPacket );
	}
	else
	{
		int nTotalLen = pPacket->HeaderLength() + pPacket->BodyLength();
	
		ASSERT( strlen(szTargetIP) > 0 && nTargetPort != 0 );

		char *pBuffer = new char[nTotalLen];

		memcpy( pBuffer, pPacket->RawHeader(), pPacket->HeaderLength() );
		memcpy( pBuffer + pPacket->HeaderLength(), pPacket->BodyData(), pPacket->BodyLength() );
		int nsend = m_pIckarSocket->SendTo( pBuffer, nTotalLen, nTargetPort, szTargetIP );

		delete [] pBuffer;

		if( nsend < 0 && !m_pIckarSocket->GetLastWouldBlockError())
			return false;

		return true;
	}
}

bool CNpUdpSession::SendNpPacket_M6V2(USHORT cmmd,
									  ULONG trid,
									  USHORT usBodyDataLength,
									  LPCSTR szBodyData,
									  LPC_NP_UTF8 pu8_To,
									  LPC_NP_UTF8 pu8_From,
									  USHORT usSessionId,
									  LPCSTR szTargetIP,
									  int nTargetPort)
{
	DLOG("CNpUdpSession::SendNpPacket_M6V2(cmm=[0x%04x] %s \n"
		"\t trid=%d, usBodyDataLength=%d, szBodyData=%p\n"
		"\t pu8_To=[%s], pu8_From=[%s], usSessionId=%d)", 
		cmmd, ::get_np_cmmd_str(cmmd), 
		trid, usBodyDataLength, szBodyData, 
		pu8_To ? (LPCSTR)pu8_To : "none", 
		pu8_From ? (LPCSTR)pu8_From : "none", usSessionId);
	
	BIN_DUMP( szBodyData, usBodyDataLength );
	
	if( IsConnectMode() )
	{
		//ASSERT( (szTargetIP==NULL || strlen(szTargetIP) == 0 ) && nTargetPort == 0 );
		
		return CNpPacketSender::SendNpPacket_M6V2( m_pIckarSocket, cmmd,
									  trid,
									  usBodyDataLength,
									  szBodyData,
									  0,
									  pu8_To,
									  pu8_From,
									  usSessionId);
	}
	else
	{
		ASSERT( strlen(szTargetIP) > 0 && nTargetPort != 0 );

		char send_buffer[DGRAM_SEND_BUFFER_DEFAULT_LEN];

		if( !CNpPacket::sMakeM6V2Header( send_buffer, 
			DGRAM_SEND_BUFFER_DEFAULT_LEN, 
			cmmd, 
			trid, 
			usBodyDataLength, 
			0,
			usSessionId,
			pu8_To, 
			pu8_From ))
			return false;
	
		M6V2_HEADER * pM6V2 = (M6V2_HEADER *)send_buffer;

		ASSERT( pM6V2->usBodyLength == usBodyDataLength );

		int nsend = 0;
		int nTotalLen = pM6V2->usHeaderLength + usBodyDataLength;
		if( nTotalLen <= DGRAM_SEND_BUFFER_DEFAULT_LEN )
		{
			memcpy( send_buffer + pM6V2->usHeaderLength, szBodyData, usBodyDataLength );

			nsend = m_pIckarSocket->SendTo( send_buffer, nTotalLen, nTargetPort, szTargetIP );
		}
		else
		{
			char * pNewBuffer = new char[nTotalLen];
			memcpy( pNewBuffer, send_buffer, pM6V2->usHeaderLength );		
			memcpy( pNewBuffer + pM6V2->usHeaderLength, szBodyData, usBodyDataLength );

			nsend = m_pIckarSocket->SendTo( pNewBuffer, nTotalLen, nTargetPort, szTargetIP );

			delete [] pNewBuffer;
		}
		
		if( nsend < 0 && !m_pIckarSocket->GetLastWouldBlockError())
			return false;
		return true;
	}	
}

bool CNpUdpSession::SendNpPacket_M6V2(const M6V2_HEADER & header, USHORT usBodyDataLength, LPCSTR szBodyData, LPC_NP_UTF8 pu8_To, LPC_NP_UTF8 pu8_From, LPCSTR szTargetIP, int nTargetPort)
{
	DLOG("CNpUdpSession::SendNpPacket_M6V2(pu8_To=[%s], pu8_From=[%s], usBodyDataLength=%d)\n", 		
		pu8_To ? (LPCSTR)pu8_To : "none",
		pu8_From ? (LPCSTR)pu8_From : "none",
		usBodyDataLength );
	
	M6V2_HEADER_DUMP( &header );
	BIN_DUMP( szBodyData, usBodyDataLength );
	
	if( IsConnectMode() )
	{
		ASSERT( (szTargetIP==NULL || strlen(szTargetIP) == 0 ) && nTargetPort == 0 );
		
		return CNpPacketSender::SendNpPacket_M6V2( m_pIckarSocket, header, usBodyDataLength, szBodyData, pu8_To, pu8_From );
	}
	else
	{
		ASSERT( strlen(szTargetIP) > 0 && nTargetPort != 0 );

		char send_buffer[DGRAM_SEND_BUFFER_DEFAULT_LEN];
		memcpy( send_buffer, &header, sizeof(M6V2_HEADER) );

		int nRouteInfoLen = CNpPacket::sMakeHeaderRouteInfo( send_buffer + sizeof(M6V2_HEADER), 
			DGRAM_SEND_BUFFER_DEFAULT_LEN - sizeof(M6V2_HEADER),
			pu8_To, 
			pu8_From );

		if( nRouteInfoLen < 0 )
			return false;
	
		M6V2_HEADER * pM6V2 = (M6V2_HEADER *)send_buffer;
		pM6V2->usHeaderLength = sizeof(M6V2_HEADER) + nRouteInfoLen;
		pM6V2->usBodyLength = usBodyDataLength;

		int nsend = 0;
		int nTotalLen = pM6V2->usHeaderLength + usBodyDataLength;
		if( nTotalLen <= DGRAM_SEND_BUFFER_DEFAULT_LEN )
		{
			memcpy( send_buffer + pM6V2->usHeaderLength, szBodyData, usBodyDataLength );

			nsend = m_pIckarSocket->SendTo( send_buffer, nTotalLen, nTargetPort, szTargetIP );
		}
		else
		{
			char * pNewBuffer = new char[nTotalLen];
			memcpy( pNewBuffer, send_buffer, pM6V2->usHeaderLength );		
			memcpy( pNewBuffer + pM6V2->usHeaderLength, szBodyData, usBodyDataLength );

			nsend = m_pIckarSocket->SendTo( pNewBuffer, nTotalLen, nTargetPort, szTargetIP );

			delete [] pNewBuffer;
		}
		
		if( nsend < 0 && !m_pIckarSocket->GetLastWouldBlockError())
			return false;
		return true;
	}
}
	

bool CNpUdpSession::SendNpPacket_MediaData(USHORT trid, USHORT usBodyDataLength, LPCSTR szBodyData, LPCSTR szTargetIP, int nTargetPort)
{
	DLOG("[ENTER] CNpUdpSession::SendPacket_MediaData(trid=%d, usBodyDataLength=%d, szBodyData=%p)"
		", IsConnectMode()=%d, target=%s:%d",
		trid, usBodyDataLength, szBodyData,
		IsConnectMode(),
		szTargetIP ? szTargetIP : "", nTargetPort );
	
	ASSERT( (trid & 0xF000) == 0 );	// ֻ 4bit  trid  ؼ ȵ..
	if( (trid & 0xF000) != 0 )
		return false;
	
	//	BIN_DUMP( szBodyData, usBodyDataLength );
	
	if( IsConnectMode() )
	{
		ASSERT( (szTargetIP==NULL || strlen(szTargetIP) == 0 ) && nTargetPort == 0 );

		CNpPacketSender::SendNpPacket_MediaData( m_pIckarSocket, trid, usBodyDataLength, szBodyData );
	}
	else
	{
		ASSERT( strlen(szTargetIP) > 0 && nTargetPort != 0 );
		
		char *pPacket = new char[MEDIA_PACKET_HEADER_SIZE + usBodyDataLength];

		memcpy( pPacket+1, &trid, 1 );
		memcpy( pPacket, ((char*)&trid) + 1, 1 );
		pPacket[0] |= 0xC0; // ֻ 3bit  110  ..
/*
		if( !(pPacket[0] & 0x80) )
		{
			int a=0;
		}
*/
		memcpy( pPacket + 2, &usBodyDataLength, 2 );
		memcpy( pPacket + MEDIA_PACKET_HEADER_SIZE, szBodyData, usBodyDataLength );


		m_pIckarSocket->SendTo( pPacket, MEDIA_PACKET_HEADER_SIZE + usBodyDataLength, nTargetPort, szTargetIP );

		delete [] pPacket;
	}
		
	DLOG("[LEAVE] CNpUdpSession::SendPacket_MediaData(trid=%d, usBodyDataLength=%d, szBodyData=%p)\n",
		trid, usBodyDataLength, szBodyData);
	
	return true;
}

bool CNpUdpSession::SendNpPacket_MediaPing(USHORT trid, LPCSTR szTargetIP, int nTargetPort)
{
	DLOG("[ENTER] CNpUdpSession::SendNpPacket_MediaPing(trid=%d), IsConnectMode()=%d, target=%s:%d", trid, 
		IsConnectMode(),
		szTargetIP ? szTargetIP : "", nTargetPort );
	
	ASSERT( (trid & 0xF000) == 0 );	// ֻ 4bit  trid  ؼ ȵ..
	if( (trid & 0xF000) != 0 )
		return false;
	
	if( IsConnectMode() )
	{
		ASSERT( (szTargetIP==NULL || strlen(szTargetIP) == 0 ) && nTargetPort == 0 );

		CNpPacketSender::SendNpPacket_MediaPing( m_pIckarSocket, trid );
	}
	else
	{
		ASSERT( strlen(szTargetIP) > 0 && nTargetPort != 0 );

		char pPacket[MEDIA_PACKET_HEADER_SIZE];

		memcpy( pPacket+1, &trid, 1 );
		memcpy( pPacket, ((char*)&trid) + 1, 1 );
		pPacket[0] |= 0xE0; // ֻ 3bit  111  ..

		memset( pPacket+2, 0, 2 );		// body length  0 .

		m_pIckarSocket->SendTo( pPacket, MEDIA_PACKET_HEADER_SIZE, nTargetPort, szTargetIP );
	}
	
	DLOG("[LEAVE] CNpUdpSession::SendNpPacket_MediaPing(trid=%d)\n", trid);
	
	return true;
}

bool CNpUdpSession::SendNpPacket_SessionControlPacket(UCHAR data, LPCSTR szTargetIP, int nTargetPort)
{
	DLOG("[ENTER] CNpUdpSession::SendNpPacket_SessionControlPacket(data=%x)", data);
	
	data |= 0x80;
	ASSERT( (data & 0xE0) == 0x80 );	// ֻ 3bit [100]  ݵ  Ǿ ..
	if( (data & 0xE0) != 0x80 )
		return false;
	
	if( IsConnectMode() )
	{
		ASSERT( (szTargetIP==NULL || strlen(szTargetIP) == 0 ) && nTargetPort == 0 );
		m_pIckarSocket->Send( (char*)&data, 1 );
	}
	else
	{
		ASSERT( strlen(szTargetIP) > 0 && nTargetPort != 0 );
		m_pIckarSocket->SendTo( (char*)&data, 1, nTargetPort, szTargetIP );
	}
	
	DLOG("[LEAVE] CNpUdpSession::SendNpPacket_SessionControlPacket(data=%x)\n", data);
	
	return true;
}


void CNpUdpSession::SendRawData(LPSTR pBuffer, int nDataLen)
{
	if( m_pIckarSocket )
	{
		m_pIckarSocket->Send(pBuffer, nDataLen);
	}
}

int CNpUdpSession::RecvRawData(LPSTR pBuffer, int nDataLen)
{
	if( m_pIckarSocket )
	{
		return m_pIckarSocket->Receive(pBuffer, nDataLen);
	}
	return 0;
}


void CNpUdpSession::Delete()
{
	delete this;
}



///////////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////////
// receive events : _IIckarSocketEvents
// 
void CALLBACK CNpUdpSession::OnIckarSocket_Connect(IIckarSocket * pIckarSocket, int nSocketError)
{
	//ASSERT( false );
}

void CALLBACK CNpUdpSession::OnIckarSocket_Close(IIckarSocket * pIckarSocket, int nSocketError)
{
	DLOG("[ENTER] CNpUdpSession::OnIckarSocket_Close(pIckarSocket=%p, nError=%d), this=%p", pIckarSocket, nSocketError, this );
	
	if(m_pEvents)
		m_pEvents->OnNpUdpSession_Error( this, nSocketError, false );
	
	DLOG("[LEAVE] CNpUdpSession::OnIckarSocket_Close(pIckarSocket=%p, nError=%d), this=%p", pIckarSocket, nSocketError, this );
}	

void CALLBACK CNpUdpSession::OnIckarSocket_Accept(IIckarSocket * pIckarSocket, int nSocketError)
{
	DLOG("[ENTER] CNpUdpSession::OnIckarSocket_Accept(pIckarSocket=%p, nError=%d), this=%p", pIckarSocket, nSocketError, this );
	
	ASSERT( false );
	
	DLOG("[LEAVE] CNpUdpSession::OnIckarSocket_Accept(pIckarSocket=%p, nError=%d), this=%p", pIckarSocket, nSocketError, this );
}

void CALLBACK CNpUdpSession::OnIckarSocket_SendReady(IIckarSocket * pIckarSocket)
{

}

void CALLBACK CNpUdpSession::OnIckarSocket_ReceiveReady(IIckarSocket * pIckarSocket, int nBytes)
{
	ASSERT( nBytes > 0 );
#define MAX_UDP_PACKET_LENGTH		2048
	
	char szRecvBuff[MAX_UDP_PACKET_LENGTH+1];		// UDP  MTU ٸ ũ ȴ.
	char szFromIP[16];
	szFromIP[0] = 0;
	USHORT usFromPort = 0;
	
	int n;
	if( IsConnectMode() )
	{
		// In using bsd socket and nonblocking, don't need to use recvfrom... (recvfrom() may be returned 0)
		n = m_pIckarSocket->Receive( szRecvBuff, MAX_UDP_PACKET_LENGTH );
	}
	else
	{
		n = m_pIckarSocket->ReceiveFrom( szRecvBuff, MAX_UDP_PACKET_LENGTH, usFromPort, szFromIP );
	}
	
	if( n <= 0 )
	{
		int nError = m_pIckarSocket->GetLastError();
		if( n < 0 
#if defined(WIN32)
		   && WSAEWOULDBLOCK == nError
#else
		   && EWOULDBLOCK == nError
#endif
		   )
			return;
		
		if(m_pEvents)
			m_pEvents->OnNpUdpSession_Error( this, nError, false );
		return;
	}
	szRecvBuff[n] = 0;
	
	CNpPacket npPacket;
	
	int nHeaderLength = CNpPacket::sParseNpPacketHeader(&npPacket, szRecvBuff, n);
	if( nHeaderLength < 1 )
	{
		if(m_pEvents)
			m_pEvents->OnNpUdpSession_Error( this, 0, true );
	}
	else
	{
		npPacket.m_pBodyData = szRecvBuff + npPacket.HeaderLength();
		if(m_pEvents)
			m_pEvents->OnNpUdpSession_RecvPacket( this, &npPacket, szFromIP, usFromPort );
	}
}
